/*

	ruucpd V 1.0 (C) 2000 by Florian Lohoff <flo@rfc822.org> 

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

*/

#define _DAEMON_PORT 513
#define _DAEMON_NAME "ruucpd"
#define _PATH_UUCICO "/usr/sbin/uucico"

#include <syslog.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <errno.h>
#include <signal.h>
#include <wait.h>
#include <pty.h>
#include <unistd.h>
#include <string.h>

#include <tcpd.h>


int go2back(void ) {
	int fd;
	/* Go into Background */
	switch(fork()) {
		case -1:
			return(-1);
		case 0:
			break;
		default:
			_exit(0);
	}

	fd=open("/dev/null", O_RDWR);

	if (fd == -1) 
		return(-1);	

	dup2(fd, 0);
	dup2(fd, 1);
	dup2(fd, 2);

	return(0);
}

/*
** Setup the listening socket.
*/

static int setup_listen(int listen_port, u_long listen_addr) {
	const int one = 1;
	int listenfd, ret;
	struct sockaddr_in sl_sin;

	memset(&sl_sin, 0, sizeof(sl_sin));

	sl_sin.sin_family = AF_INET;
	sl_sin.sin_port = htons(listen_port);
	sl_sin.sin_addr.s_addr = listen_addr;

	listenfd = socket(PF_INET, SOCK_STREAM, 0);

	if (listenfd == -1) {
		printf("Error: socket: %s\n", strerror(errno));
		return(-1);
	}

	ret = setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

	if (ret == -1) {
		printf("Error: setsockopt: %s\n", strerror(errno));
		return(-1);
	}

	ret = bind(listenfd, (struct sockaddr *) &sl_sin, sizeof(struct sockaddr));
	
	if (ret == -1) {
		printf("Error: bind: %s\n", strerror(errno));
		return(-1);
	}

	if (listen(listenfd, 3) == -1) {
		printf("Error: listen: %s\n", strerror(errno));
		return(-1);
	}

	return (listenfd);
}

static void
getstr(int sockfd, char *buf, int cnt, const char *errmsg)
{
	char c;

	do {
		if (read(sockfd, &c, 1) != 1)
			exit(1);
		if (--cnt < 0)
			syslog(LOG_CRIT, errmsg, 0);
		*buf++ = c;
	} while (c != 0);
}

int process_client(int sockfd) {
	struct request_info	ri;

	char	rusername[255],
		lusername[255],
		term[64],
		c,
		*pos;

	openlog(_DAEMON_NAME, LOG_PID | LOG_CONS, LOG_DAEMON);

	/* Run tcp wrapper */
	request_init(&ri, RQ_FILE, sockfd, RQ_DAEMON, _DAEMON_NAME, NULL);
	fromhost(&ri);
	if (!hosts_access(&ri)) {
		syslog(LOG_ERR, "connection denied from %s", eval_client(&ri));
		return(-1);
	}
	syslog(LOG_ERR, "connection from %s", eval_client(&ri));

	/* Read a zero byte */
	read(sockfd, &c, 1);

	/* Get logins informations */
	getstr(sockfd, rusername, sizeof(rusername), "remuser too long");
	getstr(sockfd, lusername, sizeof(lusername), "locuser too long");
	getstr(sockfd, term, sizeof(term), "Terminal type too long");

	/* Ack the data */
	c = 0x0;
	write(sockfd, &c, 1);

	/* Find realm seperator */
	pos = strchr(lusername, '@');	

	/* And set the string termination there */
	if (pos != NULL) {
		*pos=0x0;
	} else {
		syslog(LOG_ERR, "No realm seperator found in %s\n", lusername);
		return(-1);
	}

	/* Log them */
	syslog(LOG_DEBUG, "running uucico for system %s", lusername);

	dup2(sockfd, 0);
	dup2(sockfd, 1);
	dup2(sockfd, 2);

	execl(_PATH_UUCICO, "uucico", "-u",
		lusername, 0);

	/* Not reached */
	return(0);
}

/*
 * ** Handle SIGCHLD.
 * */

static void sigchld(int sig) {
	while (waitpid(-1, &sig, WNOHANG) > 0) ;

	signal(SIGCHLD, sigchld);
}

int main(void ) {
	int connectfd,
		listenfd,
		len = sizeof(struct sockaddr);

	struct sockaddr cliaddr;

	listenfd=setup_listen(_DAEMON_PORT, htonl(INADDR_ANY));

	if (go2back() == -1) {
		printf("Couldnt go to background\n");
		exit(-1);
	}

	openlog(_DAEMON_NAME, LOG_PID | LOG_CONS, LOG_DAEMON);

	signal(SIGCHLD, sigchld);

	if (listenfd != -1) {
		for(;;) {
			connectfd = accept(listenfd, &cliaddr, &len);	

			if (connectfd == -1) 
				continue;

			if (!fork()) {
				close(listenfd);
				process_client(connectfd);
				close(connectfd);
				exit(0);
			}
			close(connectfd);
		}
	}
	return(0);
}
